'use strict';
/* globals dS */

(function ($) {

    var ajaxRequestPost = function (surl, data, successCallback, errorCallback, completeCallback, beforeSendCallback) {
        var result = null;
        var isAsync = !!successCallback;

        var request = $.ajax({
            url: surl,
            type: 'POST',
            async: isAsync,
            contentType: 'application/json',
            beforeSend: function (xhr) {
                var csrf = dS.utils.getCookie('csrf-token');
                if (csrf !== '') {
                    xhr.setRequestHeader('X-Csrf-Token', csrf);
                }
                if (typeof beforeSendCallback === 'function') {
                    beforeSendCallback();
                }
            },
            data: JSON.stringify(data),
            success: successCallback,
            error: errorCallback,
            complete: completeCallback
        });

        return (isAsync ? request : result);
    };

    // shortcuts
    var i18n;
    var ajaxRequest;
    var callout;
    var copyTableBody;

    // "global" variables
    var vdcImplementationId = 'HomeConnect_Container';
    var vdcToken = null;
    var deviceStateDescriptions = {};
    var dsmuids = [];
    var vdcDsuid;

    var SECOND = 1000;
    var reregisterDelay = 10 * SECOND;
    var vdcPresentTimeout = 5 * SECOND;
    var stateRetryDelay = 2 * SECOND;
    var stateRetryLimit = 5;

    var authProviders = {
        physicalHomeAppliances: 'siemens',
        simulatedHomeAppliances: 'siemensSimulator'
    };
    var authProvider = authProviders.physicalHomeAppliances;

    var cloudAvailable = null;

    var $deviceTable;
    var $deviceTableWorkingCopy;
    var $rediscoverButton;
    var $searchingForDevicesPopup;
    var $remoteConnectivityPopup;
    var $loginFailedPopup;
    var $noControllersPopup;
    var $noDevicesFoundPopup;
    var $oAuthButton;
    var $oAuthStatus;
    var homeConnect = {
        status: 'disconnected',
        statusNames: {
            disconnected: 'Not logged in',
            connected: 'Connected',
            offline: 'Offline',
            'dS internal error': 'dS internal error'
        }
    };

    function getParameterByName (name) {
        var regexS = '[?&]' + name + '=([^&#]*)';
        var regex = new RegExp(regexS);
        var results = regex.exec(window.location.search);
        if (results == null) {
            return '';
        } else {
            return decodeURIComponent(results[1].replace(/\+/g, ' '));
        }
    }

    /**
     * CONNECTING TO HOME CONNECT VIA oAuth
     */

    function updateRowsState ($row, status, needsDeviceStatusUpdate) {
        switch (status) {
            case 'dS internal error':
                $row.find('.actions .button').setState('disabled');
                break;
            case 'disconnected':
                $row.addClass('parent-absent');
                $row.find('.actions .button').setState('disabled');
                $row.find('.status').html('<span class="rel">' + dS.utils.faWarning + ' <span>' + i18n._('Device status could not be determined at the moment') + '</span></span>');
                break;
            case 'offline':
            case 'connected':
                $row.removeClass('parent-absent');
                if (needsDeviceStatusUpdate && !$row.hasClass('.absent')) {
                    $row.find('.actions .button').setState('default');
                    $row.find('.status .rel').setState('in progress');
                    queryForStateDescriptions($row.attr('data-dsuid'));
                }
                break;
            default:
                break;
        }
    }

    function oAuthUpdateStatus (status, initial) {
        if (status === homeConnect.status && !initial) {
            return;
        }

        // update device table rows - class, button state
        if (!initial) {
            var $rows = $deviceTable.find('tr[data-dsuid]');
            var needsDeviceStatusUpdate = (status === 'offline' || status === 'connected') && homeConnect.status !== 'offline' && homeConnect.status !== 'connected';
            $rows.each(function () {
                updateRowsState($(this), status, needsDeviceStatusUpdate);
            });
        }

        // update log in/out button
        $oAuthButton.text(i18n._(status === 'disconnected' || status === 'dS internal error' ? 'Log in to Home-Connect account' : 'Disconnect from Home-Connect account'));
        $oAuthButton.setState(status === 'dS internal error' ? 'halted' : 'default');

        // update status text
        $oAuthStatus.find('.status').html((status !== 'connected' ? dS.utils.faWarning + '&nbsp;' : '') + i18n._(homeConnect.statusNames[status]));

        // update rediscover button
        if (status !== 'connected') {
            $rediscoverButton.setState('disabled');
        } else if (initial) {
            $rediscoverButton.setState('enabled');
        }

        homeConnect.status = status;
    }

    function oAuthCheckStatus (initial) {
        if (initial) {
            $oAuthStatus = $('<p>' + i18n._('Home-Connect account:') + ' <em class="status inline">' + i18n._('checking status...') + '</em></p>');
            $('#home_connect').append($oAuthStatus);
            $oAuthButton.setState('in progress');
        }

        if (typeof vdcDsuid === 'undefined') {
            oAuthUpdateStatus('disconnected', initial);
            $oAuthButton.setState('disabled');
            return;
        }

        ajaxRequest('/json/property/vdcquery', {
            query: 'getProperty(deviceId=' + vdcDsuid + ')/homeConnectAccountStatus(*)'
        }, function (data) {
            if (data && data.ok && data.result && data.result.elements && data.result.elements[0] && typeof data.result.elements[0].value !== 'undefined') {
                oAuthUpdateStatus(data.result.elements[0].value, initial);
            } else {
                oAuthUpdateStatus('dS internal error', initial);
            }
            checkRemoteConnect(true);
        }, function () {
            oAuthUpdateStatus('dS internal error', initial);
            checkRemoteConnect(true);
        });

        // check if the user tried to log in and failed
        var qsParams = dS.utils.getQueryStringParams();
        var keys = Object.keys(qsParams);
        if (keys.indexOf('addon_final_page') > -1 && keys.indexOf('vdc_dsuid') > -1 && qsParams.message === 'Internal error') {
            // show popup
            $loginFailedPopup.open();
        }
    }

    function oAuthLogIn () {
        $oAuthButton.setState('in progress');

        function postErrorCallback (e) {
            callout('error', i18n._('Login to Home-Connect account has failed. Please try again.'), e);
            $oAuthButton.setState('enabled');
        }

        ajaxRequestPost('/api/v1/apartment/controllers/' + vdcDsuid + '/authorize', {
            authScope: 'IdentifyAppliance Monitor Control Settings',
            authResponseType: 'code',
            provider: authProvider,
            lang: i18n.languageSuffix.replace('_', '-'),
            redirectUri: window.location.href.split('?')[0]
        }, function (data) {
            if (data.location) {
                window.location.replace(data.location);
            } else {
                postErrorCallback([ { label: i18n._('Response'), content: JSON.stringify(data) } ]);
            }
        }, function (error) {
            var errorDetails = [
                { label: i18n._('Response code'), content: error.status },
                { label: i18n._('Status'), content: error.statusText }
            ];
            if (error.responseText) {
                errorDetails.push({ label: i18n._('Response'), content: error.responseText });
            }
            postErrorCallback(errorDetails);
        }, function () {
            $oAuthButton.setState('default');
        });
    }

    function oAuthLogOut () {
        // send oAuth request with empty credentials
        $oAuthButton.setState('in progress');

        ajaxRequest(' /json/circuit/storeAccessToken', {
            dsuid: vdcDsuid,
            authData: '',   // '{ access_token: "", refresh_token: "" }',
            authScope: ''
        }, function (data) {
            if (data.ok) {
                oAuthUpdateStatus('disconnected');
                buildTable();
            } else {
                callout('error', i18n._('Logging out from Home Connect failed. Please try again in a few minutes.'));
            }
        }, function () {
            callout('error', i18n._('Logging out from Home Connect failed. Please try again in a few minutes.'));
        }, function () {
            $oAuthButton.setState('enabled');
        });
    }

    function oAuthButtonHandler () {
        if (homeConnect.status === 'disconnected') {
            oAuthLogIn();
        } else {
            oAuthLogOut();
        }
    }

    /**
     * SET UP THE APP
     */

    function queryForVdc () {
        ajaxRequest('/json/apartment/getVdcs', {
            implementationId: vdcImplementationId
        }, function (data) {
            if (data && data.result && data.result.vdcs && data.result.vdcs[0] && typeof data.result.vdcs[0].dSUID !== 'undefined') {
                $noControllersPopup.close();
                vdcDsuid = data.result.vdcs[0].dSUID;

                oAuthCheckStatus(true);
            } else {
                $noControllersPopup.open();
                $oAuthButton.setState('disabled');
                $rediscoverButton.setState('halted');

                setTimeout(queryForVdc, vdcPresentTimeout);
            }
        }, function () {
            callout('error', i18n._('Checking for controllers has failed. Please refresh the page in a few minutes to try again.'));
            oAuthCheckStatus(true);
        }, function () {
            $rediscoverButton.setState('in progress');
            $deviceTable.setTableState('in progress');
        });
    }

    function showAuthProviderSelector () {
        var $apiSelector = $('<select id="DeviceApiSelection">');
        for (var provider in authProviders) {
            if (Object.prototype.hasOwnProperty.call(authProviders, provider)) {
                $apiSelector.append($('<option value="' + authProviders[provider] +
                '" ' + (authProviders[provider] === authProvider ? 'selected' : '') + '>' +
                authProviders[provider] + ' (' + provider + ')</option>'));
            }
        }
        $('body').prepend($apiSelector);

        $apiSelector.on('change', function () {
            var $selectedOption = $(this).find(':selected');
            authProvider = $selectedOption.val();
        });
    }

    function feedDetectedCallback () {
        if (getParameterByName('mode') === 'expert') {
            showAuthProviderSelector();
        }
    }

    function buildEnvironment () {
        i18n = dS.i18n;
        ajaxRequest = dS.utils.ajaxRequest;
        callout = dS.utils.showStatusCallout;
        copyTableBody = dS.utils.copyTableBody;

        $deviceTable = $('#Devices');
        $deviceTableWorkingCopy = $deviceTable.clone();
        $rediscoverButton = $('#button_rediscover_devices');
        $searchingForDevicesPopup = $('#SearchingForDevices');
        $noDevicesFoundPopup = $('#NoDevicesFound');
        $noControllersPopup = $('#NoControllersFound');
        $oAuthButton = $('#oauth_button');
        $remoteConnectivityPopup = $('#RemoteConnectivity');
        $loginFailedPopup = $('#LoginFailed');

        dS.utils.detectFeed(feedDetectedCallback);

        queryForVdc();
    }

    function translate () {
        i18n.buildUpLang([ 'outdated_browser_fragment', 'content_fragment', 'missing_remote_connectivity', 'bottom_fragment' ]);
        $('body').addClass('loaded');
    }

    /**
     * DEVICES LISTING
     */

    function queryForDeviceState (dsuid, retrying) {
        var $stateCell = $deviceTable.find('.state_' + dsuid);

        function retry (retrying) {
            setTimeout(function () {
                queryForDeviceState(dsuid, retrying);
            }, stateRetryDelay);
        }

        function errorCallback () {
            if (retrying >= stateRetryLimit) {
                $stateCell
                    .setState('halted')
                    .html(dS.utils.faWarning + ' <span>' + i18n._('Device connected but no data available from device') + '</span>');
            } else {
                retry(retrying ? retrying + 1 : 1);
            }
        }

        ajaxRequest('/json/device/getInfoOperational', {
            dsuid: dsuid
        }, function (data) {
            if (data && data.result && typeof data.result.states !== 'undefined') {
                data = data.result.states;
                var sString = '';
                var value;
                var title;
                var option;
                var optionSuccessfullyRead = true;
                for (var state in data) {
                    if (Object.prototype.hasOwnProperty.call(data, state) && typeof data[state].value !== 'undefined') {
                        value = data[state].value;
                        if (deviceStateDescriptions && deviceStateDescriptions[dsuid] && deviceStateDescriptions[dsuid][state] &&
                            typeof deviceStateDescriptions[dsuid][state].options !== 'undefined') {
                            title = deviceStateDescriptions[dsuid][state].title;
                            option = deviceStateDescriptions[dsuid][state].options[value];
                            optionSuccessfullyRead = optionSuccessfullyRead && !!option;
                            sString += '<tr>' +
                                '           <td>' + (title ? title + ':' : '') + '</td>' +
                                '           <td class="state"><em>' + (option ? option : '?') + '</em></td>' +
                                '       </tr>';
                        } else {
                            optionSuccessfullyRead = false;
                        }
                    }
                }
                if (sString) {
                    sString = '<table>' + sString + '</table>';
                }
                $stateCell.html(sString);
                if (retrying >= stateRetryLimit || optionSuccessfullyRead) {
                    $stateCell.setState('halted');
                } else {
                    retry(retrying ? retrying + 1 : 1);
                }
            } else {
                errorCallback();
            }
        }, errorCallback);
    }

    function queryForStateDescriptions (dsuid) {
        ajaxRequest('/json/device/getInfo', {
            dsuid: dsuid,
            lang: i18n.languageSuffix,
            filter: 'stateDesc'
        }, function (data) {
            if (data && data.result && typeof data.result.stateDescriptions !== 'undefined') {
                deviceStateDescriptions[dsuid] = data.result.stateDescriptions;
            }

            queryForDeviceState(dsuid);
        }, function () {
            $('.state_' + dsuid)
                .setState('halted')
                .html(dS.utils.faWarning + ' <span>' + i18n._('Device connected but no data available from device') + '</span>');
        });
    }

    function insertDevice (device) {
        var noConnection = homeConnect.status === 'dS internal error' || homeConnect.status === 'disconnected';

        var feed = dS.utils.feed === 'fieldtest'
            ? 'production'
            : dS.utils.feed;
        var configLink = device.productInfo.ConfigLink +
            '?dsuid=' + device.dSUID +
            '&gtin=' + device.productInfo.EAN +
            '&lang=' + i18n.languageSuffix +
            '&feed=' + feed +
            '&context=overview' +
            '&AppName=Configurator' +
            (vdcToken !== null ? '&token=' + vdcToken : '');
        var disabled = (noConnection || !device.present || !device.productInfo.ConfigLink ? ' disabled' : '');
        configLink = '<a' +
            (device.productInfo.ConfigLink ? ' href="' + configLink + '"' : '') +
            ' class="button secondary recheck-remote-connect' + disabled + '"' +
            ' target="_blank"' + disabled + '>' +
            i18n._('open device settings page') + '</a>';

        var status = '<span class="rel in-progress">&nbsp;</span>';     // first time only
        if (noConnection) {
            status = '<span class="rel">' + dS.utils.faWarning + ' <span>' + i18n._('Device status could not be determined at the moment') + '</span></span>';
        } else if (!device.present) {
            status = '<span class="rel">' + dS.utils.faWarning + ' <span>' + i18n._('Device is absent') + '</span></span>';
        } else {
            var $currentState = $deviceTable.find('[data-dsuid="' + device.dSUID + '"] .status .rel');
            if ($currentState.length) {
                status = '<span class="rel in progress">' + $currentState.html() + '</span>';
            }
        }

        var $tr = $('<tr>');
        $tr
            .attr('data-dsuid', device.dSUID)
            .addClass(device.present ? '' : 'absent')
            .addClass(noConnection ? 'parent-absent' : '')
            .html(' <td data-before="' + i18n._('Name') + '" class="name">' + (device.name ? device.name : '-') + '</td>' +
                '   <td data-before="' + i18n._('Model') + '">' + device.properties.HardwareInfo + '</td>' +
                '   <td data-before="' + i18n._('Status') + '" class="status state_' + device.dSUID + '">' + status + '</td>' +
                '   <td class="actions text-center large-text-right">' + configLink + '</td>');   // don't put whitespace here before </td> (it'll mess up :empty styling)

        $deviceTableWorkingCopy.append($tr);
    }

    function insertDevices (devices) {
        dsmuids = [];
        for (var dsuid in devices) {
            if (Object.prototype.hasOwnProperty.call(devices, dsuid)) {
                insertDevice(devices[dsuid]);
                if (dsmuids.indexOf(devices[dsuid].DSMeterDSUID) === -1) {
                    dsmuids.push(devices[dsuid].DSMeterDSUID);
                }
            }
        }

        $deviceTable.setTableState('success');
        copyTableBody($deviceTableWorkingCopy, $deviceTable);
        $searchingForDevicesPopup.close();

        if (homeConnect.status !== 'dS internal error') {
            for (dsuid in devices) {
                if (Object.prototype.hasOwnProperty.call(devices, dsuid)) {
                    if (devices[dsuid].present) {
                        queryForStateDescriptions(dsuid);
                    }
                }
            }
        }
    }

    function queryForDevices () {

        function errorCallback () {
            if (homeConnect.status === 'connected') {
                $rediscoverButton.setState('enabled');
            } else {
                $rediscoverButton.setState('default').setState('disabled');
            }
            $deviceTable.setTableState('error');
            $searchingForDevicesPopup.close();
        }

        ajaxRequest('/json/property/query2', {
            query: '/apartment/zones/zone0/devices/*(dSUID,name,DSMeterDSUID,present)/*(*)/*(*)'
        }, function (data) {
            if (data.ok && data.result) {
                data = data.result;
                // skip non-Siemens devices
                for (var dsuid in data) {
                    if (!data[dsuid].productInfo ||
                        !data[dsuid].productInfo.EAN ||
                        data[dsuid].DSMeterDSUID !== vdcDsuid) {
                        delete data[dsuid];
                    }
                }
                if (Object.keys(data).length) {
                    insertDevices(data);
                } else {
                    $deviceTable.setTableState('empty');
                    $searchingForDevicesPopup.close();
                    if (homeConnect.status !== 'disconnected' && homeConnect.status !== 'dS internal error') {
                        $noDevicesFoundPopup.open();
                    }
                }

                if (homeConnect.status === 'connected') {
                    $rediscoverButton.setState('enabled');
                } else {
                    $rediscoverButton.setState('default').setState('disabled');
                }
            } else {
                errorCallback();
            }
        }, errorCallback);
    }

    function buildTable (initial) {
        copyTableBody($deviceTable, $deviceTableWorkingCopy);
        $deviceTableWorkingCopy.find('[data-dsuid]').remove();
        $deviceTable.setTableState('in progress');
        if (initial !== true) {
            $rediscoverButton.setState('in progress');    // different behaviour between first page load and refresh button click
        }

        setTimeout(function () {
            // dS.utils.abortRequests();    // @TODO - make requests cancellable by criteria
            queryForDevices();
        }, 0);
    }

    /**
     * CLOUD CONNECTION
     */

    function isCloudAvailable (fRefresh) {
        if (!fRefresh && cloudAvailable !== null) {
            return cloudAvailable;
        }

        cloudAvailable = false;
        var result = ajaxRequest('/cgi-bin/remoteaccessinfo.sh', {});
        if (result && result.success && result.data && typeof result.data.remoteaccess !== 'undefined') {
            cloudAvailable = result.data.remoteaccess;
        }
        return cloudAvailable;
    }

    function reregister () {
        if (homeConnect.status !== 'connected') {
            return;
        }
        $rediscoverButton.setState('in progress');
        $searchingForDevicesPopup.open();
        copyTableBody($deviceTable, $deviceTableWorkingCopy);
        $deviceTable.setTableState('in progress');

        for (var i = 0, len = dsmuids.length; i < len; i++) {
            ajaxRequest('/dsa/json/dsm/restartRegistration', { dsuid: dsmuids[i] }, true);
        }

        setTimeout(buildTable, reregisterDelay);
    }

    function checkRemoteConnect (initial) {
        vdcToken = null;
        if (!isCloudAvailable(true)) {
            $remoteConnectivityPopup.open();
            $deviceTable.setTableState('disconnected');
            $rediscoverButton.setState('disabled');
            $oAuthButton.setState('enabled');
        } else {
            ajaxRequest('/json/property/query2', {
                query: '/system/security/vdcToken(value)'
            }, function (tempVdcToken) {
                tempVdcToken = tempVdcToken.result;
                if (tempVdcToken.vdcToken && typeof tempVdcToken.vdcToken.value !== 'undefined') {
                    vdcToken = tempVdcToken.vdcToken.value;
                }

                buildTable(initial);
            }, function () {
                buildTable(initial);
            });
        }
    }

    function reCheckRemoteConnect (e) {
        if (vdcToken === null) {
            e.preventDefault();
            checkRemoteConnect();
        }
    }

    /**
     * INITIALIZE EVENTS
     */

    function initializeEvents () {
        $(document)
            .on('click', '.recheck-remote-connect', reCheckRemoteConnect)
            .on('click', '#button_rediscover_devices', reregister)
            .on('click', '#oauth_button', oAuthButtonHandler);
    }

    $(document).ready(function () {
        buildEnvironment();
        translate();

        initializeEvents();
    });

})(jQuery);
